#!/usr/bin/env python
# -*- coding: utf-8 -*-
import copy
import time
import os

import numpy as np
import dearpygui.dearpygui as dpg

from openvi.node_editor.util import dpg_get_value, dpg_set_value

from openvi.node.node_abc import DpgNodeABC
from openvi.node_editor.util import convert_cv_to_dpg

from openvi.node.deep_learning_node.object_detection.YOLOX.yolox import YOLOX

from openvi.node.deep_learning_node.object_detection.coco_class_names import (
    coco_class_names,
)
from openvi.node.deep_learning_node.object_detection.coco_class_names_only_person import (
    coco_class_names_only_person,
)

from openvi.node.draw_node.draw_util.draw_util import draw_object_detection_info


class Node(DpgNodeABC):
    _ver = "0.0.1"

    node_label = "Object Detection"
    node_tag = "ObjectDetection"

    _min_val = 0.0
    _max_val = 1.0

    _opencv_setting_dict = None

    # モデル設定
    _model_class = {
        "YOLOX-Nano(416x416)": YOLOX,
    }
    _model_base_path = (
        os.path.dirname(os.path.abspath(__file__)) + "/object_detection/"
    )
    _model_path_setting = {
        "YOLOX-Nano(416x416)": _model_base_path
        + "YOLOX/model/yolox_nano.onnx",
    }
    _model_class_name_list = {
        "YOLOX-Nano(416x416)": coco_class_names,
    }

    _model_instance = {}

    def __init__(self):
        pass

    def add_node(
        self,
        parent,
        node_id,
        pos=[0, 0],
        opencv_setting_dict=None,
        callback=None,
    ):
        # タグ名
        tag_node_name = str(node_id) + ":" + self.node_tag
        tag_node_input01_name = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":Input01"
        )
        tag_node_input01_value_name = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":Input01Value"
        )
        tag_node_input02_name = (
            tag_node_name + ":" + self.TYPE_TEXT + ":Input02"
        )
        tag_node_input02_value_name = (
            tag_node_name + ":" + self.TYPE_TEXT + ":Input02Value"
        )
        tag_node_input03_name = (
            tag_node_name + ":" + self.TYPE_FLOAT + ":Input03"
        )
        tag_node_input03_value_name = (
            tag_node_name + ":" + self.TYPE_FLOAT + ":Input03Value"
        )
        tag_node_output01_name = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":Output01"
        )
        tag_node_output01_value_name = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":Output01Value"
        )
        tag_node_output02_name = (
            tag_node_name + ":" + self.TYPE_TIME_MS + ":Output02"
        )
        tag_node_output02_value_name = (
            tag_node_name + ":" + self.TYPE_TIME_MS + ":Output02Value"
        )

        tag_provider_select_name = (
            tag_node_name + ":" + self.TYPE_TEXT + ":Provider"
        )
        tag_provider_select_value_name = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":ProviderValue"
        )

        # OpenCV向け設定
        self._opencv_setting_dict = opencv_setting_dict
        small_window_w = self._opencv_setting_dict["process_width"]
        small_window_h = self._opencv_setting_dict["process_height"]
        use_pref_counter = self._opencv_setting_dict["use_pref_counter"]
        use_gpu = self._opencv_setting_dict["use_gpu"]

        # 初期化用黒画像
        black_image = np.zeros((small_window_w, small_window_h, 3))
        black_texture = convert_cv_to_dpg(
            black_image,
            small_window_w,
            small_window_h,
        )

        # テクスチャ登録
        with dpg.texture_registry(show=False):
            dpg.add_raw_texture(
                small_window_w,
                small_window_h,
                black_texture,
                tag=tag_node_output01_value_name,
                format=dpg.mvFormat_Float_rgba,
            )

        # ノード
        with dpg.node(
            tag=tag_node_name,
            parent=parent,
            label=self.node_label,
            pos=pos,
        ):
            # 入力端子
            with dpg.node_attribute(
                tag=tag_node_input01_name,
                attribute_type=dpg.mvNode_Attr_Input,
            ):
                dpg.add_text(
                    tag=tag_node_input01_value_name,
                    default_value="Input BGR image",
                )
            # 画像
            with dpg.node_attribute(
                tag=tag_node_output01_name,
                attribute_type=dpg.mvNode_Attr_Output,
            ):
                dpg.add_image(tag_node_output01_value_name)
            # 使用アルゴリズム
            with dpg.node_attribute(
                tag=tag_node_input02_name,
                attribute_type=dpg.mvNode_Attr_Static,
            ):
                dpg.add_combo(
                    list(self._model_class.keys()),
                    default_value=list(self._model_class.keys())[0],
                    width=small_window_w,
                    tag=tag_node_input02_value_name,
                )
            if use_gpu:
                # CPU/GPU切り替え
                with dpg.node_attribute(
                    tag=tag_provider_select_name,
                    attribute_type=dpg.mvNode_Attr_Static,
                ):
                    dpg.add_radio_button(
                        ("CPU", "GPU"),
                        tag=tag_provider_select_value_name,
                        default_value="CPU",
                        horizontal=True,
                    )
            # スコア閾値
            with dpg.node_attribute(
                tag=tag_node_input03_name,
                attribute_type=dpg.mvNode_Attr_Input,
            ):
                dpg.add_slider_float(
                    tag=tag_node_input03_value_name,
                    label="score",
                    width=small_window_w - 80,
                    default_value=0.3,
                    min_value=self._min_val,
                    max_value=self._max_val,
                    callback=None,
                )
            # 処理時間
            if use_pref_counter:
                with dpg.node_attribute(
                    tag=tag_node_output02_name,
                    attribute_type=dpg.mvNode_Attr_Output,
                ):
                    dpg.add_text(
                        tag=tag_node_output02_value_name,
                        default_value="elapsed time(ms)",
                    )

        return tag_node_name

    def update(
        self,
        node_id,
        connection_list,
        node_image_dict,
        node_result_dict,
    ):
        tag_node_name = str(node_id) + ":" + self.node_tag
        input_value02_tag = (
            tag_node_name + ":" + self.TYPE_TEXT + ":Input02Value"
        )
        input_value03_tag = (
            tag_node_name + ":" + self.TYPE_FLOAT + ":Input03Value"
        )
        output_value01_tag = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":Output01Value"
        )
        output_value02_tag = (
            tag_node_name + ":" + self.TYPE_TIME_MS + ":Output02Value"
        )

        tag_provider_select_value_name = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":ProviderValue"
        )

        small_window_w = self._opencv_setting_dict["process_width"]
        small_window_h = self._opencv_setting_dict["process_height"]
        use_pref_counter = self._opencv_setting_dict["use_pref_counter"]
        use_gpu = self._opencv_setting_dict["use_gpu"]

        # 接続情報確認
        connection_info_src = ""
        for connection_info in connection_list:
            connection_type = connection_info[0].split(":")[2]
            if connection_type == self.TYPE_FLOAT:
                # 接続タグ取得
                source_tag = connection_info[0] + "Value"
                destination_tag = connection_info[1] + "Value"
                # 値更新
                input_value = round(float(dpg_get_value(source_tag)), 3)
                input_value = max([self._min_val, input_value])
                input_value = min([self._max_val, input_value])
                dpg_set_value(destination_tag, input_value)
            if connection_type == self.TYPE_IMAGE:
                # 画像取得元のノード名(ID付き)を取得
                connection_info_src = connection_info[0]
                connection_info_src = connection_info_src.split(":")[:2]
                connection_info_src = ":".join(connection_info_src)

        # 画像取得
        frame = node_image_dict.get(connection_info_src, None)

        # スコア閾値
        score_th = round(float(dpg_get_value(input_value03_tag)), 3)

        # CPU/GPU選択状態取得
        provider = "CPU"
        if use_gpu:
            provider = dpg_get_value(tag_provider_select_value_name)

        # モデル情報取得
        model_name = dpg_get_value(input_value02_tag)
        model_path = self._model_path_setting[model_name]
        model_class = self._model_class[model_name]
        class_name_dict = self._model_class_name_list[model_name]

        model_name_with_provider = model_name + "_" + provider

        # モデル取得
        if frame is not None:
            if model_name_with_provider not in self._model_instance:
                if provider == "CPU":
                    providers = ["CPUExecutionProvider"]
                    self._model_instance[
                        model_name_with_provider
                    ] = model_class(
                        model_path,
                        providers=providers,
                    )
                else:
                    self._model_instance[
                        model_name_with_provider
                    ] = model_class(model_path)

        # 計測開始
        if frame is not None and use_pref_counter:
            start_time = time.perf_counter()

        result = {}
        if frame is not None:
            bboxes, scores, class_ids = self._model_instance[
                model_name_with_provider
            ](frame)
            if len(bboxes) > 0:
                result["bboxes"] = bboxes.tolist()
                result["scores"] = scores.tolist()
                result["class_ids"] = class_ids.tolist()
                result["class_names"] = class_name_dict
                result["score_th"] = score_th
            else:
                result["bboxes"] = []
                result["scores"] = []
                result["class_ids"] = []
                result["class_names"] = class_name_dict
                result["score_th"] = score_th

        # 計測終了
        if frame is not None and use_pref_counter:
            elapsed_time = time.perf_counter() - start_time
            elapsed_time = int(elapsed_time * 1000)
            dpg_set_value(
                output_value02_tag, str(elapsed_time).zfill(4) + "ms"
            )

        # 描画
        if frame is not None:
            debug_frame = copy.deepcopy(frame)
            debug_frame = draw_object_detection_info(
                debug_frame,
                score_th,
                bboxes,
                scores,
                class_ids,
                class_name_dict,
            )
            texture = convert_cv_to_dpg(
                debug_frame,
                small_window_w,
                small_window_h,
            )
            dpg_set_value(output_value01_tag, texture)

        return frame, result

    def close(self, node_id):
        pass

    def get_setting_dict(self, node_id):
        tag_node_name = str(node_id) + ":" + self.node_tag
        input_value02_tag = (
            tag_node_name + ":" + self.TYPE_TEXT + ":Input02Value"
        )
        input_value03_tag = (
            tag_node_name + ":" + self.TYPE_FLOAT + ":Input03Value"
        )

        # 選択モデル
        model_name = dpg_get_value(input_value02_tag)
        # スコア閾値
        score_th = round(float(dpg_get_value(input_value03_tag)), 3)

        pos = dpg.get_item_pos(tag_node_name)

        setting_dict = {}
        setting_dict["ver"] = self._ver
        setting_dict["pos"] = pos
        setting_dict[input_value02_tag] = model_name
        setting_dict[input_value03_tag] = score_th

        return setting_dict

    def set_setting_dict(self, node_id, setting_dict):
        tag_node_name = str(node_id) + ":" + self.node_tag
        input_value02_tag = (
            tag_node_name + ":" + self.TYPE_TEXT + ":Input02Value"
        )
        input_value03_tag = (
            tag_node_name + ":" + self.TYPE_FLOAT + ":Input03Value"
        )

        model_name = setting_dict[input_value02_tag]
        score_th = setting_dict[input_value03_tag]

        dpg_set_value(input_value02_tag, model_name)
        dpg_set_value(input_value03_tag, score_th)
