#!/usr/bin/env python
# -*- coding: utf-8 -*-
import time
import numpy as np
import dearpygui.dearpygui as dpg


from openvi.node_editor.util import dpg_get_value, dpg_set_value

from openvi.node.node_abc import DpgNodeABC
from openvi.node_editor.util import convert_cv_to_dpg


class Node(DpgNodeABC):
    _ver = "0.0.1"

    node_label = "WebCam"
    node_tag = "WebCam"

    _opencv_setting_dict = None

    def __init__(self):
        pass

    def add_node(
        self,
        parent,
        node_id,
        pos=[0, 0],
        opencv_setting_dict=None,
        callback=None,
    ):
        # タグ名
        tag_node_name = str(node_id) + ":" + self.node_tag
        tag_node_input01_name = (
            tag_node_name + ":" + self.TYPE_INT + ":Input01"
        )
        tag_node_input01_value_name = (
            tag_node_name + ":" + self.TYPE_INT + ":Input01Value"
        )
        tag_node_output01_name = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":Output01"
        )
        tag_node_output01_value_name = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":Output01Value"
        )
        tag_node_output02_name = (
            tag_node_name + ":" + self.TYPE_TIME_MS + ":Output02"
        )
        tag_node_output02_value_name = (
            tag_node_name + ":" + self.TYPE_TIME_MS + ":Output02Value"
        )

        # OpenCV向け設定
        self._opencv_setting_dict = opencv_setting_dict
        small_window_w = self._opencv_setting_dict["input_window_width"]
        small_window_h = self._opencv_setting_dict["input_window_height"]
        device_no_list = self._opencv_setting_dict["device_no_list"]
        use_pref_counter = self._opencv_setting_dict["use_pref_counter"]

        # 初期化用黒画像
        black_image = np.zeros((small_window_w, small_window_h, 3))
        black_texture = convert_cv_to_dpg(
            black_image,
            small_window_w,
            small_window_h,
        )

        # テクスチャ登録
        with dpg.texture_registry(show=False):
            dpg.add_raw_texture(
                small_window_w,
                small_window_h,
                black_texture,
                tag=tag_node_output01_value_name,
                format=dpg.mvFormat_Float_rgba,
            )

        # ノード
        with dpg.node(
            tag=tag_node_name,
            parent=parent,
            label=self.node_label,
            pos=pos,
        ):
            # カメラNo選択 コンボボックス
            with dpg.node_attribute(
                tag=tag_node_input01_name,
                attribute_type=dpg.mvNode_Attr_Static,
            ):
                dpg.add_combo(
                    device_no_list,
                    width=small_window_w - 100,
                    label="Device No",
                    tag=tag_node_input01_value_name,
                )
            # カメラ画像
            with dpg.node_attribute(
                tag=tag_node_output01_name,
                attribute_type=dpg.mvNode_Attr_Output,
            ):
                dpg.add_image(tag_node_output01_value_name)
            # 処理時間
            if use_pref_counter:
                with dpg.node_attribute(
                    tag=tag_node_output02_name,
                    attribute_type=dpg.mvNode_Attr_Output,
                ):
                    dpg.add_text(
                        tag=tag_node_output02_value_name,
                        default_value="elapsed time(ms)",
                    )

        return tag_node_name

    def update(
        self,
        node_id,
        connection_list,
        node_image_dict,
        node_result_dict,
    ):
        tag_node_name = str(node_id) + ":" + self.node_tag
        input_value01_tag = (
            tag_node_name + ":" + self.TYPE_INT + ":Input01Value"
        )
        output_value01_tag = (
            tag_node_name + ":" + self.TYPE_IMAGE + ":Output01Value"
        )
        output_value02_tag = (
            tag_node_name + ":" + self.TYPE_TIME_MS + ":Output02Value"
        )

        device_no_list = self._opencv_setting_dict["device_no_list"]
        camera_capture_list = self._opencv_setting_dict["camera_capture_list"]
        small_window_w = self._opencv_setting_dict["input_window_width"]
        small_window_h = self._opencv_setting_dict["input_window_height"]
        use_pref_counter = self._opencv_setting_dict["use_pref_counter"]

        # カメラNo取得
        camera_no = dpg_get_value(input_value01_tag)

        # VideoCapture()インスタンス取得
        camera_capture = None
        if camera_no != "":
            camera_no = int(camera_no)
            camera_index = device_no_list.index(camera_no)
            camera_capture = camera_capture_list[camera_index]

        # 計測開始
        if camera_no != "" and use_pref_counter:
            start_time = time.perf_counter()

        # 画像取得
        frame = None
        if camera_capture is not None:
            ret, frame = camera_capture.read()
            if not ret:
                return

        # 計測終了
        if camera_no != "" and use_pref_counter:
            elapsed_time = time.perf_counter() - start_time
            elapsed_time = int(elapsed_time * 1000)
            dpg_set_value(
                output_value02_tag, str(elapsed_time).zfill(4) + "ms"
            )

        # 描画
        if frame is not None:
            texture = convert_cv_to_dpg(
                frame,
                small_window_w,
                small_window_h,
            )
            dpg_set_value(output_value01_tag, texture)

        return frame, None

    def close(self, node_id):
        pass

    def get_setting_dict(self, node_id):
        tag_node_name = str(node_id) + ":" + self.node_tag

        pos = dpg.get_item_pos(tag_node_name)

        setting_dict = {}
        setting_dict["ver"] = self._ver
        setting_dict["pos"] = pos

        return setting_dict

    def set_setting_dict(self, node_id, setting_dict):
        pass
